<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2019 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2018.3/1751405
 */

namespace Api\Controller;

use Api\AbstractApiController;
use P4\Exception;
use Users\Authentication\Helper;
use Zend\Http\Response;
use Zend\View\Model\JsonModel;

/**
 * Swarm Login API
 *
 * @SWG\Resource(
 *   apiVersion="v9",
 *   basePath="/api/v9/"
 * )
 */
class LoginController extends AbstractApiController
{
    const RESULTS = 'results';
    const METHOD  = 'method';
    const METHODS = 'methods';
    const CODE    = 'code';
    const OPTION  = 'option';
    const ERROR   = 'error';
    const TOKEN   = 'token';
    const POLL    = 'poll';
    const PROMPT  = 'prompt';
    const TRIGGER = 'trigger';

    const NEXTSTATE  = 'nextState';
    const SUCCESSMSG = 'successMsg';
    const MFASTATUS  = 'mfaStatus';
    const CHALLENGE  = 'challenge';

    const RESPONSEMETHOD = 'responseMethod';

    const MFASUCCESS     = "mfaSuccess";
    const MFAPOLLPENDING = "mfaPending";
    const MFAFAILED      = "mfaFailed";
    const CHECK_AUTH     = 'check-auth';

    // Session values
    const COOKIE_LIFETIME            = 'cookie_lifetime';
    const REMEMBERED_COOKIE_LIFETIME = 'remembered_cookie_lifetime';

     /**
     * @SWG\Api(
     *     path="listmethods/",
     *     @SWG\Operation(
     *         method="GET",
     *         summary="Get List of 2FA Methods",
     *         notes="Returns the complete list of methods of 2FA.",
     *         nickname="listMethods"
     *     )
     * )
     *
     * @apiUsageExample Listing 2FA Methods
     *
     *   To list the 2FA methods:
     *
     *   ```bash
     *   curl -u "username:password" \
     *        "https://myswarm.url/api/v9/login/listmethods"
     *   ```
     *
     *   Swarm responds with a list of 2FA methods:
     *
     *   ```json
     *   {
     *     "results": {
     *       "methods": {
     *         "1": {
     *           "methodName": "Method Name will be here",
     *           "methodDesc": "Method Description will be here"
     *         },
     *         "2": {
     *           "methodName": "Method Name will be here",
     *           "methodDesc": "Method Description will be here"
     *         },
     *         "3": {
     *           "methodName": "Method Name will be here",
     *           "methodDesc": "Method Description will be here"
     *         },
     *         "4": {
     *           "methodName": "Method Name will be here",
     *           "methodDesc": "Method Description will be here"
     *         }
     *       }
     *     },
     *     "option": {
     *       "persist": "option",
     *       "nextState": "init-auth"
     *     },
     *     "code": 200
     *   }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *     {
     *       "results": {
     *         "methods": {
     *           "1": {
     *             "methodName": "Method Name will be here",
     *             "methodDesc": "Method Description will be here"
     *           },
     *           "2": {
     *             "methodName": "Method Name will be here",
     *             "methodDesc": "Method Description will be here"
     *           },
     *           "3": {
     *             "methodName": "Method Name will be here",
     *             "methodDesc": "Method Description will be here"
     *           },
     *           "4": {
     *             "methodName": "Method Name will be here",
     *             "methodDesc": "Method Description will be here"
     *           }
     *         }
     *       },
     *       "option": {
     *         "persist": "option",
     *         "nextState": "init-auth"
     *       },
     *       "code": 200
     *     }
     *
     * @return mixed
     */
    public function listMethodsAction()
    {
        $services = $this->getServiceLocator();
        $p4User   = $services->get('p4_user');

        // Set the flags to get list of methods
        $flags = array('-S', 'list-methods');

        try {
            // Run the command to fetch the data
            $commandResults = $p4User->run('login2', $flags, null, true);
            $commandData    = $commandResults->getData();
            $isMfa          = sizeof($commandData) > 1;
            $option         = $isMfa ? $commandData[1] : null;
            // If we don't get returned data[1] then we can assume the command failed.
            if (!$option) {
                if ($this->isMfaCommandError($commandData[0])) {
                    return new JsonModel(
                        array(
                            static::CODE    => 200,
                            static::RESULTS => $commandData[0]
                        )
                    );
                } else {
                    throw new Exception("Missing required data.");
                }
            }
            // Collect methods and option
            $options = $option;
            $methods = $commandData[0];

            // If we are returned 1 or more methods return them to the page.
            if (count($methods) > 0) {
                $results[static::RESULTS][static::METHODS] = $this->listMethodSort($methods);
            }

            // Set the option and code vaules to be returned.
            $results[static::OPTION] = $options;
            $results[static::CODE]   = Response::STATUS_CODE_200;

            return $this->prepareSuccessModel($results);
        } catch (Exception $e) {
            // set message to the error from P4 and just return it.
            $message = $e->getMessage();
        }

        // If we hit here then we have failed and should return the error if we have one.
        return new JsonModel(
            array(
                'isValid'     => false,
                static::CODE  => Response::STATUS_CODE_401,
                static::ERROR => $message
            )
        );
    }

    /**
     * Tests to see if the command result is due to MFA not in use.
     * @param string $commandResult the command result
     * @return bool true if the command specifies MFA is not in use
     */
    private function isMfaCommandError($commandResult)
    {
        $matches = array();
        preg_match('/does not use \bsecond\b|\bmulti\b factor authentication./', $commandResult, $matches);
        return sizeof($matches) > 0;
    }

    /**
     * @SWG\Api(
     *     path="initauth/",
     *     @SWG\Operation(
     *         method="POST",
     *         summary="initiating the 2FA authentication",
     *         notes="initiating the 2FA authentication",
     *         nickname="initAuth",
     *         @SWG\Parameter(
     *             name="method",
     *             description="The Method in which you want to use.",
     *             paramType="form",
     *             type="string",
     *             required=true
     *         )
     *     )
     * )
     *
     * @apiUsageExample Initiating the 2FA authentication
     *
     *   To Initiate the user 2FA login:
     *
     *   ```bash
     *   curl -X POST -u "username:password" \
     *        -d "method=METHOD" \
     *        "https://myswarm.url/api/v9/login/initauth"
     *   ```
     *
     *   Assuming that the authenticated user has permission, Swarm responds with the next step in the 2FA login:
     *
     *   ```json
     *   {
     *     "results": {
     *       "trigger": "TriggerName",
     *       "successMsg": "Message from Authentication method"
     *     },
     *     "option": {
     *       "prompt": true,
     *       "nextState": "check-auth"
     *     },
     *     "code": 200
     *   }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *     {
     *       "results": {
     *          "trigger": "TriggerName",
     *          "successMsg": "Message from Authentication method"
     *       },
     *       "option": {
     *         "prompt": true,
     *         "nextState": "check-auth"
     *       },
     *       "code": 200
     *     }
     *
     * @param POST
     * @return JsonModel
     */
    public function initAuthAction()
    {
        $request   = $this->getRequest();
        $services  = $this->getServiceLocator();
        $p4User    = $services->get('p4_user');
        $method2fa = $request->getPost(static::METHOD);

        // Set the flags to init auth method
        $flags = array('-S', 'init-auth', '-m', $method2fa);

        try {
            // Run the command to fetch the data
            $commandResults = $p4User->run('login2', $flags, null, true);
            $initAuthData   = $commandResults->getData(0);
            if (!is_array($initAuthData)
                && $this->isMfaCommandError($initAuthData)) {
                return new JsonModel(
                    array(
                        static::CODE    => 200,
                        static::RESULTS => $initAuthData
                    )
                );
            }

            // Set the trigger and successmsg that we need to display.
            $results[static::RESULTS] = array(
                static::TRIGGER    => $initAuthData[static::TRIGGER],
                static::SUCCESSMSG => $initAuthData[static::SUCCESSMSG]
            );

            if (isset($initAuthData[static::CHALLENGE])) {
                $results[static::RESULTS][static::CHALLENGE] = $initAuthData[static::CHALLENGE];
            }

            // The option that tells us if we are prompt or poll
            $results[static::OPTION] = array(
                static::RESPONSEMETHOD => $initAuthData[static::RESPONSEMETHOD],
                static::NEXTSTATE      => $initAuthData[static::NEXTSTATE]
            );

            // Set the code to successful.
            $results[static::CODE] = Response::STATUS_CODE_200;

            return $this->prepareSuccessModel($results);
        } catch (Exception $e) {
            // set message to the error from P4 and just return it.
            $message = $e->getMessage();
        }
        // If we hit here then we have failed and should return the error if we have one.
        return new JsonModel(
            array(
                'isValid'     => false,
                static::CODE  => Response::STATUS_CODE_401,
                static::ERROR => $message
            )
        );
    }

    /**
     * @SWG\Api(
     *     path="checkauth/",
     *     @SWG\Operation(
     *         method="POST",
     *         summary="checking the 2FA authentication",
     *         notes="checking the 2FA authentication",
     *         nickname="checkAuth",
     *         @SWG\Parameter(
     *             name="token",
     *             description="The token from the user for their 2FA prompt.",
     *             paramType="form",
     *             type="string",
     *             required=true
     *         )
     *     )
     * )
     *
     * @apiUsageExample Checking the 2FA authentication
     *
     *   To Check User prompt input for 2FA
     *
     *   ```bash
     *   curl -X POST -u "username:password" \
     *        -d "token=TOKEN" \
     *        "https://myswarm.url/api/v9/login/checkauth"
     *   ```
     *
     *   Assuming that the authenticated user has permission, Swarm responds with the next step in the 2FA login:
     *
     *   ```json
     *   {
     *     "results": {
     *       "trigger": "GAuth says yes!",
     *       "successMsg": "Second factor authentication approved."
     *     },
     *     "code": 200
     *   }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *     {
     *       "results": {
     *         "trigger": "otp-generated|||GAuth says yes!",
     *         "successMsg": "Second factor authentication approved."
     *       },
     *       "code": 200
     *     }
     *
     * @param POST
     * @return JsonModel
     */
    public function checkAuthAction()
    {
        $request  = $this->getRequest();
        $services = $this->getServiceLocator();
        $p4User   = $services->get('p4_user');
        $prompt   = $request->getPOST(static::TOKEN);

        // Set the flags to check auth
        $flags = array('-S', static::CHECK_AUTH);

        if ($prompt !== null) {
            try {
                // Run the command to fetch the data
                $commandResults = $p4User->run('login2', $flags, $prompt, true);
                $checkAuthData  = $commandResults->getData();

                // Ensure the data has been set.
                if ($checkAuthData === false) {
                    $checkAuthData = $commandResults->getData(0);
                }
                // Set the code and results to be returned.
                $results[static::RESULTS]   = $checkAuthData;
                $results[static::CODE]      = Response::STATUS_CODE_200;
                $results[static::MFASTATUS] = static::MFASUCCESS;

                return $this->prepareSuccessModel($results);
            } catch (Exception $e) {
                // set message to the error from P4 and just return it.
                $message = $e->getMessage();
            }
        } else {
            try {
                $commandResults = $p4User->run('login2', $flags, $prompt, true);
                $checkAuthData  = $commandResults->getData();
                $message        = $checkAuthData[0];
                if ($this->isMfaCommandError($message)) {
                    return new JsonModel(
                        array(
                            static::CODE      => 200,
                            static::RESULTS   => $message,
                            static::MFASTATUS => static::MFASUCCESS
                        )
                    );
                }
            } catch (Exception $e) {
                // set message to the error from P4 and just return it.
                $message = $e->getMessage();
            }
        }
        // If we hit here then we have failed and should return the error if we have one.
        return new JsonModel(
            array(
                'isValid'         => false,
                static::CODE      => Response::STATUS_CODE_401,
                static::ERROR     => $message,
                static::MFASTATUS => static::MFAFAILED
            )
        );
    }

    /**
     * @SWG\Api(
     *     path="checkauth/",
     *     @SWG\Operation(
     *         method="GET",
     *         summary="checking the 2FA authentication",
     *         notes="checking the 2FA authentication",
     *         nickname="checkAuthPoll"
     *     )
     * )
     *
     * @apiUsageExample Checking the 2FA authentication
     *
     *   To Check User prompt input for 2FA
     *
     *   ```bash
     *   curl -u "username:password" \
     *        "https://myswarm.url/api/v9/login/checkauth"
     *   ```
     *
     *   Assuming that the authenticated user has permission, Swarm responds with the next step in the 2FA login:
     *
     *   ```json
     *   {
     *     "results": {
     *       "trigger": "GAuth says yes!",
     *       "successMsg": "Second factor authentication approved."
     *     },
     *     "code": 200
     *   }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *     {
     *       "results": {
     *         "trigger": "GAuth says yes!",
     *         "successMsg": "Second factor authentication approved."
     *       },
     *       "code": 200
     *     }
     *
     * @return JsonModel
     */
    public function checkAuthPollAction()
    {
        $services = $this->getServiceLocator();
        $p4User   = $services->get('p4_user');
        $message  = '';

        // Set the flags to check auth poll
        $flags = array('-S', static::CHECK_AUTH);
        try {
            // Run the command to fetch the data
            $commandResults    = $p4User->run('login2', $flags, null, true);
            $checkAuthPollData = $commandResults->getData();

            // Set the code and results to be returned.
            $results[static::RESULTS]   = $checkAuthPollData;
            $results[static::CODE]      = Response::STATUS_CODE_200;
            $results[static::MFASTATUS] = static::MFASTATUS;

            return $this->prepareSuccessModel($results);
        } catch (Exception $e) {
            // set message to the error from P4 and just return it.
            $message  = $e->getMessage();
            $pending  = (strpos($e->getMessage(), "authentication approval is pending.") > 0);
            $rejected = (strpos($e->getMessage(), "authentication was rejected.") > 0);

            if ($pending || $rejected) {
                $checkAuthPollData = array_merge($e->getResult()->getData(), $e->getResult()->getErrors());

                // Set the code and results to be returned.
                $results[static::RESULTS] = $checkAuthPollData;
                $results[static::CODE]    = Response::STATUS_CODE_200;

                if ($pending) {
                    $results[ static::MFASTATUS ] = static::MFAPOLLPENDING;
                } elseif ($rejected) {
                    $results[static::MFASTATUS] = static::MFAFAILED;
                }

                return $this->prepareSuccessModel($results);
            }
        }
        // If we hit here then we have failed and should return the error if we have one.
        return new JsonModel(
            array(
                'isValid'         => false,
                static::CODE      => Response::STATUS_CODE_401,
                static::ERROR     => $message,
                static::MFASTATUS => static::MFAFAILED
            )
        );
    }

    /**
     * @SWG\Api(
     *     path="login/",
     *     @SWG\Operation(
     *         method="POST",
     *         summary="Login to Swarm",
     *         notes="Login to Swarm",
     *         nickname="login",
     *         @SWG\ResponseMessage(code=401, message="User name or password incorrect.")
     *     )
     * )
     *
     * @apiUsageExample Logging in to swarm
     *
     *   To login:
     *
     *   ```bash
     *   curl -X POST -H "Content-Type: application/json" \
     *        -u "super:<ticket>" \
     *        -d '{"username":"swarm.user","password":"1234"}' "http://myswarm.url/api/v9/login"
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *     {
     *          "isValid": true,
     *          "messages": [],
     *           "user": {
     *              "User": "swarm.user",
     *              "FullName": "Swarm User",
     *              "Email": "swarm.user@mydomain.com",
     *              "Type": "standard",
     *              "Password": "enabled",
     *              "isAdmin": false,
     *              "isSuper": false
     *          }
     *     }
     *
     *   In the event of a failed login attempt Swarm responds with:
     *
     *   ```json
     *   {
     *      "isValid": false,
     *      "messages": ["User name or password incorrect."]
     *   }
     *   ```
     * @return JsonModel
     * @throws \Application\Config\ConfigException
     */
    public function loginAction()
    {
        // This isn't setting session data correctly so login not remembered when going to server context
        $request = $this->getRequest();
        if ($this->requestHasContentType($request, self::CONTENT_TYPE_JSON)) {
            $data = json_decode($request->getContent(), true);
        } else {
            $data = $request->getPost()->toArray();
        }
        // Clear any existing session data on login, we need to explicitly restart the session
        $services   = $this->getServiceLocator();
        $translator = $services->get('translator');
        $session    = $services->get(Helper::SESSION);
        $error      = null;
        $authUser   = null;
        $userDetail = null;
        $adapter    = null;

        $session->start();
        $session->getStorage()->clear();
        if (isset($data[Helper::USERNAME]) && isset($data[Helper::PASSWORD])) {
            $userName = $data[Helper::USERNAME];
            $password = $data[Helper::PASSWORD];
            $p4Admin  = $services->get('p4_admin');
            $candAuth = Helper::authenticateCandidates($services, $userName, $password);
            $authUser = $candAuth[Helper::AUTH_USER];
            $adapter  = $candAuth[Helper::ADAPTER];
            if ($candAuth[Helper::IS_VALID]) {
                Helper::setCookieInformation($services, $request, $data);
                $authUser = Helper::invalidateCache($authUser, $p4Admin);
            } else {
                $error = Helper::createErrorMessage($services, $this->getResponse(), $p4Admin, $userName);
            }
        } else {
            $error = $translator->t("User name and password must be provided.");
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_401);
        }
        $session->writeClose();
        return $this->buildLoginResponse($adapter, $authUser, $error);
    }

    /**
     * @SWG\Api(
     *     path="logout/",
     *     @SWG\Operation(
     *         method="POST",
     *         summary="Logout of Swarm with optional redirect",
     *         notes="Logout of Swarm",
     *         nickname="logout"
     *     )
     * )
     *
     * @apiUsageExample Logout of Swarm
     *
     *   To logout:
     *
     *   ```bash
     *   curl -X POST \
     *        -u "super:<ticket>" \
     *        "http://myswarm.url/api/v9/logout"
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 302 OK
     *
     *     {
     *          "isValid": true,
     *          "messages": []
     *     }
     *
     * @apiUsageExample Logout of Swarm without redirect
     *
     *   To logout without any redirect:
     *
     *   ```bash
     *   curl -X POST \
     *        -u "super:<ticket>" \
     *        "http://myswarm.url/api/v9/logout?stay=true"
     *   ```
     *
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *     {
     *          "isValid": true,
     *          "messages": []
     *     }
     *
     *   In the event of a failed login attempt Swarm responds with:
     *
     *   ```json
     *   {
     *      "isValid": false,
     *      "messages": ["Error message."]
     *   }
     *   ```
     *
     * @return JsonModel
     */
    public function logoutAction()
    {
        $services = $this->getServiceLocator();
        $request  = $this->getRequest();
        $error    = null;
        try {
            Helper::logout($services, $request, $this->getResponse());
        } catch (\Exception $e) {
            $error = $e->getMessage();
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_500);
        }
        return $this->buildLogoutResponse($error);
    }

    /**
     * Builds a JSON model response for logout.
     * @param mixed     $error  any error message to include.
     * @return JsonModel
     */
    private function buildLogoutResponse($error)
    {
        $response = array(
            'isValid'  => $error ? false : true,
            'messages' => $error ? array($error) : array(),
        );
        return new JsonModel($response);
    }

    /**
     * Builds a JSON response based on results.
     * @param mixed  $adapter the authentication adapter used to auth the user
     * @param string $authUser the user identifier
     * @param string $error an error (may be null)
     * @return JsonModel
     */
    private function buildLoginResponse($adapter, $authUser, $error)
    {
        $response = array(
            'isValid'  => $authUser ? true : false,
            'messages' => $error ? array($error) : array(),
        );
        if ($authUser) {
            $response = $response + array(
                    'user' =>
                        $authUser->getValuesArray() +
                        array(
                            'isAdmin' => $adapter->getUserP4()->isAdminUser(true),
                            'isSuper' => $adapter->getUserP4()->isSuperUser(),
                        )
                );
        }
        return new JsonModel($response);
    }

    /**
     * @SWG\Api(
     *     path="login/saml",
     *     @SWG\Operation(
     *         method="POST",
     *         summary="Login to Swarm with Saml",
     *         notes="Login to Swarm with Saml",
     *         nickname="samlLogin"
     *     )
     * )
     *
     * @apiUsageExample Logging in to swarm with Saml, redirects to the IDP when redirect=false or redirect is not
     * specified
     *
     * @apiSuccessExample Successful Response:
     *   HTTP/1.1 302 OK
     *
     *   {
     *      "isValid": "true"
     *   }
     *
     * @apiUsageExample Logging in to swarm with Saml returns a url from the IDP if redirect=false
     *
     * @apiSuccessExample Successful Response:
     *   HTTP/1.1 200 OK
     *
     *   {
     *      "isValid": "true",
     *      "url": "<url to redirect to>"
     *   }
     *
     */
    public function samlLoginAction()
    {
        return Helper::handleSamlLogin($this->getServiceLocator(), $this->getRequest());
    }

    /**
     * @SWG\Api(
     *     path="session",
     *     @SWG\Operation(
     *         method="GET",
     *         summary="Get the current effective user details",
     *         notes="Get user logged in",
     *         nickname="getUser"
     *     )
     * )
     *
     *
     * @apiUsageExample Getting the currently effective user details
     *
     *   To get session details:
     *
     *   ```bash
     *   curl -u "<username>:<ticket>" "http://myswarm.url/api/v9/session"
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *   HTTP/1.1 200 OK
     *
     *  {
     *      "isValid": true,
     *      "messages": [],
     *      "user": {
     *          "User": "reviewer",
     *          "FullName": "Code Reviewer",
     *          "Email": "reviewer@swarm.local",
     *          "Type": "standard",
     *          "Password": "enabled"
     *      }
     *  }
     *
     * @return mixed
     */
    public function getList()
    {
        return $this->forward(
            'Api\Controller\Users',
            null,
            null,
            array('current'=>true),
            null
        );
    }

    /**
     * @SWG\Api(
     *     path="session",
     *     @SWG\Operation(
     *         method="DELETE",
     *         summary="Destroy the current session, i.e. logout",
     *         notes="Logout of Swarm",
     *         nickname="logout"
     *     )
     * )
     *
     * @apiUsageExample Logging out of swarm
     *
     *   To login:
     *
     *   ```bash
     *   curl -u "<username>:<ticket>" -X DELETE "http://myswarm.url/api/v9/session"
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *   HTTP/1.1 200 OK
     *
     *   {
     *       "isValid": true,
     *       "messages": []
     *   }
     *
     * @return mixed|JsonModel
     */
    public function deleteList()
    {
        return $this->logoutAction();
    }

    /**
     * @SWG\Api(
     *     path="session",
     *     @SWG\Operation(
     *         method="POST",
     *         summary="Create a new Swarm session using the given credentials",
     *         notes="Login to swarm",
     *         nickname="login"
     *     )
     * )
     *
     * @apiUsageExample Logging in to Swarm
     *
     *   To login:
     *
     *   ```bash
     *   curl \
     *       -H "Content-Type: application/json" \
     *       -d '{"username": "<username>","password": "<password>","remember": "false"}' \
     *       -X POST http://localhost/api/v9/session
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *   HTTP/1.1 200 OK
     *
     *  {
     *      "isValid": true,
     *      "messages": [],
     *      "user": {
     *          "User": "reviewer",
     *          "FullName": "Code Reviewer",
     *          "Email": "reviewer@swarm.local",
     *          "Type": "standard",
     *          "Password": "enabled"
     *      }
     *  }
     *
     * @param mixed $data - ignored
     * @return mixed|JsonModel
     * @throws \Application\Config\ConfigException
     */
    public function create($data)
    {
        return $this->loginAction();
    }

    // TODO samlResponseAction
    // samlResponseAction() is here ready for when we use the API to handle the saml
    // response (when all the login actions are using the API rather than the web
    // controller). Currently it is planned that the classic login will go through
    // the web controller, not sure yet if we need it so untested for now
    /**
     * Endpoint to handle the return from the IDP to validate Saml details.
     * @return JsonModel
     */
    public function samlResponseAction()
    {
        $services = $this->getServiceLocator();
        $error    = null;
        $authUser = null;
        $p4Admin  = $services->get('p4_admin');
        $logger   = $services->get('logger');
        $session  = $services->get(Helper::SESSION);
        $session->start();
        $session->getStorage()->clear();
        $result  = Helper::handleSamlResponse($services);
        $userId  = $result[Helper::USERNAME];
        $adapter = $result[Helper::ADAPTER];
        try {
            if ($result[Helper::IS_VALID]) {
                Helper::setCookieInformation($services, $this->getRequest(), array(Helper::USERNAME => $userId));
                $authUser = Helper::invalidateCache($userId, $p4Admin);
            } else {
                $error = Helper::createErrorMessage($services, $this->getResponse(), $p4Admin, $userId);
            }
        } catch (\Exception $e) {
            $error = $e->getMessage();
            $logger->err("An error occurred using Saml login " . $error);
        }
        $session->writeClose();
        return $this->buildLoginResponse($adapter, $authUser, $error);
    }

    /**
     * Take an array of methods element sort them into separate array within an array
     *
     * @param $methods    Single dimension array
     * @return array      Multi dimension array
     */
    protected function listMethodSort($methods)
    {
        $results = array();
        $last    = 0; // Keep track of which method we last saw

        // loop though each of the data sent back and create an array to send in json
        foreach ($methods as $key => $method) {
            preg_match("/(.*)([0-9]+)$/U", $key, $spilt);
            if (isset($spilt[1]) && isset($spilt[2])) {
                $field   = $spilt[1]; // This is the name of the field
                $current = $spilt[2]; // This is the current method number we are at

                // If we are still on the same method then add the next field to it.
                if ($current === $last) {
                    $results[$current][$field] = $method;
                } else {
                    // else we set current and create the next method in the array.
                    $last = $current;

                    $results[$current][$field] = $method;
                }
            }
        }
        return $results;
    }
}
